import itertools
import sys

def return_none(): return None
def return_0(): return 0

def assert_equal(x, y):
  if x != y:
    print("------------------------------------------------------------")
    print(x)
    print("-----------------------------!=-----------------------------")
    print(y)
    print("------------------------------------------------------------")
  assert x == y

def assert_contains(element, list):
  if element not in list:
    print("------------------------------------------------------------")
    print(element)
    print("---------------------------not in---------------------------")
    print(list)
    print("------------------------------------------------------------")
  assert element in list

class SingletonValue(object):
  def __init__(self, name, id):
    self.name = name
    self.id   = id

  def __repr__(self):
    return self.name

  def __eq__(self, other):
    return isinstance(other, SingletonValue) and self.id == other.id

def flatten(list_of_lists):
  return list(itertools.chain(*list_of_lists))

def str_list(list, empty):
  if len(list) == 0: return empty
  if len(list) == 1: return str(list[0])
  return ", ".join([str(e) for e in list[:-1]]) + " and " + str(list[-1])

def str_player_list(list): return str_list(list, "nobody")

def has_method(object, method):
  return callable(getattr(object, method, None))

def matches(x, y, *, debug=False, **kwargs):
  if type(x) != type(y):
    if debug: print("Types don't match:\n  %r\n  %r" % (type(x), type(y)))
    return False
  xkeys = x.__dict__.keys()
  ykeys = y.__dict__.keys()
  if xkeys != ykeys:
    if debug: print("Keys don't match:\n  %r\n  %r" % (xkeys, ykeys))
    return False
  for k, xv, yv in [(k, x.__dict__[k], y.__dict__[k]) for k in xkeys]:
    if has_method(xv, "matches"):
      if not xv.matches(yv, **kwargs):
        if debug: print("Field %r doesn't match:\n  %r\n  %r" % (k, xv, yv))
        return False
    else:
      if xv != yv:
        if debug: print("Field %r not equal:\n  %r\n  %r" % (k, xv, yv))
        return False
  return True

def fill_randomly(x, **kwargs):
  for k, v in x.__dict__.items():
    if has_method(v, "random_instance"):
      x.__dict__[k] = v.random_instance(**kwargs)

def strict_subclasses(cls, namespace):
  """Return all strict subclasses of the given type in a namespace."""
  return [c for c in vars(namespace).values()
                  if type(c) == type
                  and issubclass(c, cls)
                  and c != cls]



from collections import *

class Phase(object):
  def resolve(self, game):
    assert game.log.current_phase == self, \
           "Use Game.resolve rather than calling resolve directly."

    self._resolve(game)

    # Remove expired effects
    for player in game.all_players:
      player.effects = [e for e in player.effects if not e.expiration.expired()]

  def _resolve(self, game):
    pass

  def resolve_actions(self, actions, *, game):
    # Resolve actions using Natural Action Resolution
    actions = sorted(actions, key=lambda action: action.precedence)
    resolved_actions = []
    while len(actions) > 0:
      dependencies = defaultdict(list)

      # Try to find an action with no dependencies
      next_action = None
      for action in actions:
        dependencies[action] = [a for a in actions if action.depends_on(a)]
        if len(dependencies[action]) == 0:
          next_action = action
          break

      # If none was found, resolve the lowest precedence action
      if not next_action:
        next_action = actions[0]

      # Resolve the action
      next_action.resolve(game)
      actions.remove(next_action)
      resolved_actions.append(next_action)

    # Post resolution. We can just do this in the same order.
    for action in resolved_actions:
      action.resolve_post(game)



from .log import *
from .factions import *
from .player import *
from .roles import Role
from .util import *

import random

EVERYONE_LOST = SingletonValue("EVERYONE_LOST", 715681895)
NO_WINNER_YET = SingletonValue("NO_WINNER_YET", 884728804)
START         = events.START

class Game(object):
  """
  A single game of Mafia.

  Usage:
    game  = Game()
    town  = game.add_faction(Town())
    mafia = game.add_faction(Mafia("Crypto Mafia"))
    alice = game.add_player("Alice", Cop(town))
    bob   = game.add_player("Bob", Doctor(town))
    eve   = game.add_player("Eve", Goon(mafia))
    game.begin()

    night0 = Night(0)
    night0.add_action(Investigate(alice, eve))
    night0.add_action(Protect(bob, alice))
    night0.add_action(FactionAction(mafia, Kill(eve, alice)))
    game.resolve(night0)

    day1 = Day(1)
    day1.set_vote(alice, eve)
    day1.set_vote(eve, alice)
    day1.set_vote(bob, eve)
    game.resolve(day1)
  """

  def __init__(self, name=None, *, seed=42):
    self.name            = name
    self.random          = random.Random(seed)
    self.log             = Log()
    self.faction_list    = []
    self.faction_names   = {}
    self.player_list     = []
    self.player_names    = {}
    self.delayed_actions = []

  def add_faction(self, faction):
    faction_name = faction.name.lower().replace(" ", "")
    assert faction_name not in self.faction_names

    faction.game = self
    self.faction_list.append(faction)
    self.faction_names[faction_name] = faction
    return faction

  def add_player(self, player, role=None, **kwargs):
    """
    Add a player to the game and return it.

    Takes either a single Player object, or a name and role from which to
    construct a Player object.
    """

    if role:
      assert isinstance(role, Role)
      player = Player(player, role, **kwargs)
    else:
      assert isinstance(player, Player)
      assert len(kwargs) == 0

    player_name = player.name.lower().replace(" ", "")
    assert player_name not in self.player_names
    assert player_name != "nobody"

    self.player_list.append(player)
    self.player_names[player_name] = player
    return player

  def begin(self):
    """Send out roles and teammates."""
    self.log.current_phase = START

    for player in self.players:
      self.log.append(events.RoleAnnouncement(player, player.role))

    for faction in self.factions:
      members = faction.members(game=self)
      if not faction.secret_membership and len(members) > 1:
        self.log.append(events.FactionAnnouncement(faction, members))
      leader = faction.leader(game=self)
      if leader and faction.action(game=self):
        self.log.append(events.FactionLeaderAnnouncement(faction, leader))

  def resolve(self, phase):
    self.log.current_phase = phase
    phase.resolve(self)

  @property
  def players(self):
    return [p for p in self.all_players if p.alive]

  @property
  def all_players(self):
    return sorted(self.player_list)

  def player_named(self, name):
    """Return a player with the given name (case insensitive) or None."""
    return self.player_names.get(name, None)

  @property
  def factions(self):
    return sorted(set([p.faction for p in self.players]))

  def faction_named(self, name):
    """Return a faction with the given name (case insensitive) or None."""
    return self.faction_names.get(name, None)

  def winners(self):
    fates = {p: p.fate(player=p, game=self) for p in self.all_players}
    winners   = sorted([f for f in fates if fates[f] == Fate.won])
    undecided = sorted([f for f in fates if fates[f] == Fate.undecided])
    losers    = sorted([f for f in fates if fates[f] == Fate.lost])

    if   len(winners)   > 0: return winners
    elif len(undecided) > 0: return NO_WINNER_YET
    else:                    return EVERYONE_LOST

  def is_game_over(self):
    winners = self.winners()
    if winners == NO_WINNER_YET: return False
    if winners == EVERYONE_LOST: return True
    return any([p.wins_exclusively for p in winners])

  def shuffled(self, list):
    """Return a deterministically shuffled version of the given list."""

    shuffled = sorted(list)
    self.random.shuffle(shuffled)
    return shuffled



from .exceptions import *
from .mixin import *

class Player(object):
  """A player in a game of Mafia."""

  def __init__(self, name, role, *, info=None):
    super().__init__()
    self.name    = name
    self.info    = info
    self.will    = ""
    self._role   = role
    self.effects = []

  def __repr__(self):
    return self.name if isinstance(self.name, str) else str(self.name)

  def __lt__(self, other):
    return self.name < other.name

  def add_effect(self, effect):
    self.effects.append(effect)

  def matches(self, other, **kwargs):
    return self == other

  @property
  def unique_name(self):
    return self.name.replace(" ", "").lower()

  @mixin("effects")
  def action(self):
    if self.role.action:
      return self.role.action.with_player(self)

  @mixin("effects")
  def action_count(self):
    return 1

  @mixin("effects")
  def alignment(self):
    return self.role.alignment

  @mixin("effects")
  def alive(self):
    return True

  @mixin("effects")
  def apparent_alignment(self):
    return self.role.apparent_alignment

  @mixin("effects")
  def apparent_factions(self):
    return self.role.apparent_factions

  @mixin("effects")
  def blocked(self):
    return False

  @mixin("effects")
  def delayed(self):
    return False

  @mixin("effects")
  def faction_action(self):
    if self.role.faction_action:
      return self.role.faction_action.with_player(self)

  @mixin("effects")
  def faction(self):
    return self.role.faction

  @mixin_fn("effects")
  def fate(self, **kwargs):
    return self.role.fate(**kwargs)

  @mixin("effects")
  def is_town_enemy(self):
    return self.role.is_town_enemy

  @mixin("effects")
  def is_town_friend(self):
    return self.role.is_town_friend

  @mixin("effects")
  def kills_visitors(self):
    return self.role.kills_visitors

  @mixin("effects")
  def must_target(self):
    return None

  @mixin_fn("effects")
  def on_killed(self, **kwargs):
    return self.role.on_killed(**kwargs)

  @mixin_fn("effects")
  def on_lynched(self, **kwargs):
    return self.role.on_lynched(**kwargs)

  @mixin("effects")
  def role(self):
    return self._role

  @mixin("effects")
  def switched_with(self):
    return self

  @mixin("effects")
  def vote_action(self):
    if self.role.vote_action:
      return self.role.vote_action.with_player(self)

  @mixin("effects")
  def votes(self):
    return self.role.votes

  @mixin("effects")
  def votes_with(self):
    return self

  @mixin("effects")
  def wins_exclusively(self):
    return self.role.wins_exclusively



from .exceptions import *
from .player import Player as RealPlayer

class Placeholder(object):
  def __str__(self):
    return self.__class__.__name__

  @property
  def switched_with(self):
    """For printing of Actions."""
    return self

class Any(Placeholder):
  def matches(self, other, **kwargs):
    return True

class Player(Placeholder):
  def matches(self, other, **kwargs):
    return isinstance(other, RealPlayer) and other.alive

  def random_instance(self, *, game, **kwargs):
    return game.random.choice(game.players)

class FactionMember(Player):
  def __init__(self, faction=None):
    self.faction = faction

  def __str__(self):
    return "%s(%s)" % (self.__class__.__name__, self.faction)

  def matches(self, other, **kwargs):
    return super().matches(other, **kwargs) and other.faction == self.faction

  def random_instance(self, *, game, **kwargs):
    return game.random.choice(self.faction.players(game.players))

class Other(Player):
  def matches(self, other, *, player, **kwargs):
    return super().matches(other) and other != player

class Self(Player):
  def matches(self, other, *, player, **kwargs):
    return other == player

class Corpse(Placeholder):
  def matches(self, other, **kwargs):
    return isinstance(other, RealPlayer) and not other.alive



from .action_base import *
from .action_helpers import *
from .effects import *
from . import events
from .log import *
from . import placeholders
from .util import *

#####################################
###   Inter-Action Dependencies   ###
#####################################

def busdrive_shares_target(self, other):
  return isinstance(other, Busdrive) and \
         len(set(self.targets).intersection(other.targets)) > 0

def roleblock_targets_player(self, other):
  return isinstance(other, Roleblock) and other.target == self.player

def protection_from_paranoid_gun_owner(self, other):
  return any([t.kills_visitors for t in self.targets]) and \
         isinstance(other, ProtectiveAction) and other.target == self.player

#############################
###   Base Action Class   ###
#############################

class Action(ActionBase):
  @property
  def dependencies(self):
    return [
      roleblock_targets_player,
      busdrive_shares_target,
      protection_from_paranoid_gun_owner,
    ]

  def __init__(self, player, targets, *, visible=True, **kwargs):
    if not isinstance(targets, list):
      targets = TargetList([targets])

    self.player      = player
    self.raw_targets = TargetList(targets)
    self.visible     = visible

  def __str__(self):
    targets = ", ".join([str(target) for target in self.targets])
    return "%s(%s, %s)" % (self.__class__.__name__, self.player, targets)

  def matches(self, other, **kwargs):
    return matches(self, other, player=self.player, **kwargs)

  @property
  def name(self):
    words = re.findall(r"[A-Z]+[a-z]*", self.__class__.__name__)
    return " ".join(words).lower()

  @property
  def targets(self):
    return [t.switched_with for t in self.raw_targets]

  @property
  def target(self):
    return self.targets[0]

  @property
  def raw_target(self):
    return self.raw_targets[0]

  def resolve(self, game):
    # Apply roleblocking
    if self.player.blocked:
      game.log.append(events.Blocked(self.player))
      return

    # Apply ventriloquisting
    if self.player.must_target:
      self.raw_targets[0] = self.player.must_target

    # Apply delaying
    if self.player.delayed:
      game.delayed_actions.append(self)
      game.log.append(events.Delayed(self.player))
      return

    # Record visit
    if self.visible:
      for target, raw_target in zip(self.targets, self.raw_targets):
        game.log.append(events.Visited(self.player, target,
                                       visible=self.player.role.visible,
                                       original_target=raw_target))
        if self.target.kills_visitors:
          resolve_kill(self.target, self.player, game=game)

    # Resolve action
    self._resolve(game)

  def resolve_post(self, game):
    if self.player.blocked or self.player.delayed: return
    self._resolve_post(game)

  def _resolve(self, game):
    pass

  def _resolve_post(self, game):
    pass

  def concrete_action(self):
    return self

  def depends_on(self, other):
    return any([d(self, other) and other != self for d in self.dependencies])

##########################
###   Action Classes   ###
##########################

class KillingAction(Action):
  pass

class ProtectiveAction(Action):
  pass

###################
###   Actions   ###
###################

class Autopsy(Action):
  precedence = 200

  def _resolve(self, game):
    visits = game.log.visits_to(self.target)
    visitors = set(v.player for v in visits if v.player != self.player)
    game.log.append(events.VisitorsResult(sorted(visitors), target=self.raw_target, to=self.player))

class Busdrive(Action):
  precedence = 0

  def __init__(self, player, target1, target2):
    super().__init__(player, [target1, target2])

  def _resolve(self, game):
    a = self.targets[0]
    b = self.targets[1]
    a.add_effect(SwitchedWith(b))
    b.add_effect(SwitchedWith(a))
    game.log.append(events.Busdriven(a, b))

class Delay(Action):
  precedence = 3

  def _resolve(self, game):
    self.target.add_effect(Delayed())

class Double(Action):
  precedence = 1000

  def _resolve(self, game):
    self.target.add_effect(ExtraAction())
    game.log.append(events.ExtraAction(self.target))

class EliteGuard(ProtectiveAction):
  precedence = 1001

  def _resolve(self, game):
    self.target.add_effect(GuardedBy(self.player, elite=True))

class Guard(ProtectiveAction):
  precedence = 1001

  def _resolve(self, game):
    self.target.add_effect(GuardedBy(self.player))

class HitmanKill(KillingAction):
  precedence = 2000

  def _resolve(self, game):
    resolve_kill(self.player, self.target, game=game, protectable=False)

class Investigate(Action):
  precedence = 1000

  def _resolve_post(self, game):
    if self.player.alive:
      game.log.append(events.InvestigationResult(self.target.apparent_alignment, target=self.raw_target, to=self.player))

class Kill(KillingAction):
  precedence = 2000

  def _resolve(self, game):
    resolve_kill(self.player, self.target, game=game)

class Pardon(Action):
  precedence = 1000

  def _resolve(self, game):
    self.target.add_effect(Unlynchable())

class Possess(Action):
  precedence = 1

  def __init__(self, player, puppet, new_target):
    super().__init__(player, puppet)
    self.new_target = new_target

  def _resolve(self, game):
    self.target.add_effect(MustTarget(self.new_target))

class Protect(ProtectiveAction):
  precedence = 1000

  def _resolve(self, game):
    self.target.add_effect(Protected())

class Roleblock(Action):
  precedence = 2

  def _resolve(self, game):
    self.target.add_effect(Blocked())

class StealVote(Action):
  precedence = 2000

  def _resolve(self, game):
    self.target.add_effect(VotesWith(self.player))

class Track(Action):
  precedence = 1000

  def _resolve_post(self, game):
    if self.player.alive:
      visits = game.log.this_phase().visits_by(self.target)
      targets = set(v.target for v in visits)
      game.log.append(events.VisiteesResult(sorted(targets), target=self.raw_target, to=self.player))

class Watch(Action):
  precedence = 1000

  def _resolve_post(self, game):
    if self.player.alive:
      visits = game.log.this_phase().visits_to(self.target)
      visitors = set(v.player for v in visits if v.player is not self.player)
      game.log.append(events.VisitorsResult(sorted(visitors), target=self.raw_target, to=self.player))



from .action_base import *
from .action_helpers import *
from .actions import *
from .alignment import *
from .day import *
from .effects import *
from . import events
from .exceptions import *
from .factions import *
from .game import *
from .log import *
from .mixin import *
from .night import *
from .parser import *
from .phase import *
from . import placeholders
from .player import *
from .roles import *
from .util import *
from .virtual_actions import *

ACTIONS = sorted(strict_subclasses(Action, mafia), key=lambda t: str(t))
FACTIONS = sorted(strict_subclasses(Faction, mafia), key=lambda t: str(t))
ROLES = sorted(set(strict_subclasses(Role, mafia)) \
             - set([ModifierRole, NeutralRole]), key=lambda t: str(t))



"""
Mixins are objects that can be "attached" to other objects to override their
functions or properties. This is used by effects to temporarily modify Player
and Role objects.

Usage:
  class Base(object):
    def __init__(self):
      self.mixins = []

    @mixin("mixins")
    def foo(self): return 123

  class Mixin(object):
    def __init__(self, value):
      self.foo = value

  base = Base()
  assert(base.foo == 123)
  base.mixins.append(Mixin(456))
  assert(base.foo == 456)
"""

from functools import wraps

def mixin_fn(mixin_attr):
  def decorator(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
      def call_next():
        for mixin in reversed(getattr(self, mixin_attr)):
          # Look for the function variant first
          if hasattr(mixin, f.__name__ + "_fn"):
            def call(mixin):
              mixin_f = getattr(mixin, f.__name__ + "_fn")
              value = mixin_f(lambda: next(next_gen)(), *args, **kwargs)
              return value
            yield lambda mixin=mixin: call(mixin)

          # Look for the attribute variant second
          if hasattr(mixin, f.__name__):
            def call(mixin):
              value = getattr(mixin, f.__name__)
              return value
            yield lambda mixin=mixin: call(mixin)

        # Fall back to the base implementation
        yield lambda: f(self, *args, **kwargs)

      next_gen = call_next()
      return next(next_gen)()

    return wrapper

  return decorator

def mixin(mixin_attr):
  return lambda f: property(mixin_fn(mixin_attr)(f))



from .alignment import *
from .player import *
from .util import *

from termcolor import colored

PUBLIC = SingletonValue("PUBLIC", 462028731)
START = SingletonValue("Start", 502402573)

class Event(object):
  def __init__(self, *, phase=None, to=None):
    if isinstance(to, Player):
      to = [to]

    self.phase = phase  # Filled in by Log.append
    self.to    = to     # None, a list of players, or PUBLIC

  def __eq__(self, other):
    return type(self) == type(other) and self.__dict__ == other.__dict__

  def __str__(self):
    if not self.to:
      return "%s: [%s]" % (self.phase, self.message)
    elif self.to == PUBLIC:
      return "%s: %s" % (self.phase, self.message)
    else:
      to = self.to if not isinstance(self.to, list) else \
           ", ".join([str(p) for p in self.to])
      return "%s: %s: %s" % (self.phase, to, self.message)

  @property
  def color(self):
    if self.to != None: return "cyan"

  @property
  def style(self):
    if self.to == PUBLIC: return ["bold"]

  def colored_str(self):
    return colored(str(self), self.color, attrs=self.style)

class Announcement(Event):
  pass

class PublicEvent(Event):
  def __init__(self, **kwargs):
    super().__init__(to=PUBLIC, **kwargs)

class Result(Event):
  def __init__(self, *, target, to, **kwargs):
    super().__init__(to=to, **kwargs)
    self.target = target

#########################
###   Announcements   ###
#########################

class FactionAnnouncement(Event):
  def __init__(self, faction, members, **kwargs):
    super().__init__(to=members, **kwargs)
    self.faction = faction
    self.members = members

  @property
  def message(self):
    return "You are %s." % self.faction.name

class FactionLeaderAnnouncement(Event):
  def __init__(self, faction, leader, **kwargs):
    super().__init__(to=leader, **kwargs)
    self.faction = faction
    self.leader  = leader

  @property
  def message(self):
    if self.phase == START:
      return "You are the leader of %s." % self.faction
    else:
      return "You are now the leader of %s." % self.faction

class RoleAnnouncement(Event):
  def __init__(self, player, role, **kwargs):
    super().__init__(to=player, **kwargs)
    self.player = player
    self.role   = role

  @property
  def message(self):
    if self.phase == START:
      return "You are the %s." % self.player.role
    else:
      return "You are now the %s." % self.player.role

##################
###   Events   ###
##################

class Blocked(Event):
  color = "yellow"

  def __init__(self, player, **kwargs):
    super().__init__(**kwargs)
    self.player = player

  @property
  def message(self):
    return "%s was blocked." % self.player

class Busdriven(Event):
  color = "yellow"

  def __init__(self, player1, player2, **kwargs):
    super().__init__(**kwargs)
    self.player1 = player1
    self.player2 = player2

  @property
  def message(self):
    return "%s was busdriven with %s." % (self.player1, self.player2)

class Delayed(Event):
  color = "yellow"

  def __init__(self, player, **kwargs):
    super().__init__(**kwargs)
    self.player = player

  @property
  def message(self):
    return "%s's action was delayed." % self.player

class Died(PublicEvent):
  color = "red"

  def __init__(self, player, **kwargs):
    super().__init__(**kwargs)
    self.player = player

  @property
  def message(self):
    return "%s, the %s, has died." % (self.player, self.player.role)

class ExtraAction(Event):
  color = "green"

  def __init__(self, player, **kwargs):
    super().__init__(to=player, **kwargs)

  @property
  def message(self):
    return "You may use your action an extra time next night."

class Lynched(Died):
  @property
  def message(self):
    return "%s, the %s, was lynched." % (self.player, self.player.role)

class NoDeaths(PublicEvent):
  color = "green"

  @property
  def message(self):
    return "Nobody died."

class NoLynch(PublicEvent):
  color = "green"

  @property
  def message(self):
    return "Nobody was lynched."

class Protected(Event):
  color = "green"

  def __init__(self, player, **kwargs):
    super().__init__(**kwargs)
    self.player = player

  @property
  def message(self):
    return "%s was protected." % self.player

class Visited(Event):
  def __init__(self, player, target, *, visible=True, original_target=None, **kwargs):
    super().__init__(**kwargs)
    self.player          = player
    self.target          = target
    self.visible         = visible
    self.original_target = original_target or self.target

  @property
  def message(self):
    visited = "visited" if self.visible else "secretly visited"
    target = self.target if self.target == self.original_target else \
             "%s (busdriven from %s)" % (self.target, self.original_target)
    return "%s %s %s." % (self.player, visited, target)

class VotedFor(Event):
  def __init__(self, player, vote, *, votes=1, original_vote=None, **kwargs):
    super().__init__(**kwargs)
    self.player        = player
    self.vote          = vote
    self.votes         = votes
    self.original_vote = original_vote or self.vote

  @property
  def message(self):
    votes = "" if self.votes == 1 else " with %d votes" % self.votes
    vote = self.vote if self.vote == self.original_vote else \
           "%s (politicianed from %s)" % (self.vote, self.original_vote)
    return "%s voted for %s%s." % (self.player, vote, votes)

###################
###   Results   ###
###################

class VisiteesResult(Result):
  def __init__(self, visitees, *, target, to, **kwargs):
    super().__init__(target=target, to=to, **kwargs)
    self.visitees = visitees

  @property
  def message(self):
    return "%s visited %s." % (self.target, str_player_list(self.visitees))

class VisitorsResult(Result):
  def __init__(self, visitors, *, target, to, **kwargs):
    super().__init__(target=target, to=to, **kwargs)
    self.visitors = visitors

  @property
  def message(self):
    return "%s visited %s." % (str_player_list(self.visitors), self.target)

class InvestigationResult(Result):
  def __init__(self, alignment, *, target, to, **kwargs):
    super().__init__(target=target, to=to, **kwargs)
    self.alignment = "good" if alignment == Alignment.good else "evil"

  @property
  def message(self):
    return "%s is %s." % (self.target, self.alignment)



from .actions import *
from .factions import *
from . import effects
from . import events
from . import placeholders
from .player import *
from .util import *
from .virtual_actions import *

import re

#############################
###   Base Role Classes   ###
#############################

class RoleBase(object):
  def __init__(self, faction):
    assert isinstance(faction, Faction)
    self._faction       = faction
    self._fake_factions = []

  @property
  def action(self):
    return None

  @property
  def adjectives(self):
    return []

  @property
  def apparent_factions(self):
    return [self.faction] + self.fake_factions

  @property
  def alignment(self):
    return self.faction.alignment

  @property
  def apparent_alignment(self):
    return self.alignment

  @property
  def descriptions(self):
    return []

  @property
  def faction(self):
    return self._faction

  @property
  def faction_action(self):
    return None

  @property
  def fake_factions(self):
    return self._fake_factions

  def fate(self, **kwargs):
    return self.faction.fate(**kwargs)

  @property
  def is_town_enemy(self):
    return self.faction.is_town_enemy

  @property
  def is_town_friend(self):
    return self.faction.is_town_friend

  @property
  def kills_visitors(self):
    return False

  def on_killed(self, *, game, player, **kwargs):
    game.log.append(events.Died(player))
    resolve_death(player, game=game)

  def on_lynched(self, *, game, player, **kwargs):
    game.log.append(events.Lynched(player))
    resolve_death(player, game=game)

  @property
  def visible(self):
    return True

  @property
  def vote_action(self):
    return None

  @property
  def votes(self):
    return 1

  @property
  def wins_exclusively(self):
    return self.faction.wins_exclusively

class Role(object):
  def __init__(self, faction_or_role):
    if isinstance(faction_or_role, Faction):
      self.base = RoleBase(faction_or_role)
    else:
      self.base = faction_or_role

  def __getattr__(self, attr):
    if attr == "base":
      raise AttributeError

    return getattr(self.base, attr)

  def __str__(self):
    return "%s %s" % (self.faction.adjective, self.adjective)

  @classmethod
  def name(cls):
    return " ".join(re.findall(r"[A-Z]+[a-z]*", cls.__name__))

  @property
  def adjective(self):
    return " ".join(self.adjectives)

  @property
  def adjectives(self):
    return [self.name()] + self.base.adjectives

  @property
  def descriptions(self):
    return [self.description] + self.base.descriptions

  @property
  def objective(self):
    return self.faction.objective

########################
###   Role Classes   ###
########################

class ModifierRole(Role):
  def __init__(self, faction_or_role):
    assert isinstance(faction_or_role, Role)
    super().__init__(faction_or_role)

class NeutralRole(Role):
  def __init__(self, role=None):
    super().__init__(role or ThirdParty("Third-Party"))

  @property
  def alignment(self):
    return Alignment.neutral

  @property
  def wins_exclusively(self):
    return False

#################
###   Roles   ###
#################

class ActionDoubler(Role):
  description = "You may double one player each night. " \
                "That player may use their action twice the following night."

  @property
  def action(self):
    return Double(placeholders.Self(), placeholders.Player())

class Bodyguard(Role):
  description = "You may guard one player each night. " \
                "If that player is killed, you die instead of them."

  @property
  def action(self):
    return Guard(placeholders.Self(), placeholders.Other())

class Bulletproof(ModifierRole):
  description = "You cannot be killed at night."

  def on_killed(self, *, game, player, protectable, **kwargs):
    if protectable:
      game.log.append(events.Protected(player))
    else:
      self.base.on_killed(game=game, player=player,
                          protectable=protectable, **kwargs)

class Busdriver(Role):
  description = "You may busdrive two players each night. " \
                "Anyone who targets the first player will automatically " \
                "target the second player instead, and vice versa. " \
                "You cannot target yourself."

  @property
  def action(self):
    return Busdrive(placeholders.Self(), placeholders.Other(), placeholders.Other())

class Cop(Role):
  description = "You may investigate one player each night. " \
                "You discover their alignment — 'Good' if they are " \
                "Town, 'Evil' if they are Mafia or Third Party."

  @property
  def action(self):
    return Investigate(placeholders.Self(), placeholders.Player())

class Delayer(Role):
  description = "You may delay one player each night. " \
                "That player's action will be resolved the following " \
                "night along with their regular action from that night."

  @property
  def action(self):
    return Delay(placeholders.Self(), placeholders.Player())

class Doctor(Role):
  description = "You may protect one player each night. " \
                "That player cannot be killed that night. " \
                "You cannot protect yourself."
  @property
  def action(self):
    return Protect(placeholders.Self(), placeholders.Other())

class DoubleVoter(Role):
  description = "Your lynch vote counts as two votes."

  @property
  def votes(self):
    return 2

class EliteBodyguard(Role):
  description = "You may guard one player each night. " \
                "If that player is killed, you die instead of them. " \
                "You also kill whoever killed them."

  @property
  def action(self):
    return EliteGuard(placeholders.Self(), placeholders.Other())

class ForensicInvestigator(Role):
  description = "You may investigate one dead player each night. " \
                "You discover everyone who ever visited that player " \
                "throughout the course of the game."

  @property
  def action(self):
    return Autopsy(placeholders.Self(), placeholders.Corpse())

class Goon(Role):
  description = "Your faction leader may order you to kill another player."

  @property
  def faction_action(self):
    return Kill(placeholders.Self(), placeholders.Player())

class Godfather(Goon):
  description = "You appear innocent to cop investigations."

  @property
  def apparent_alignment(self):
    return Alignment.good

class Governor(Role):
  description = "Whoever you vote for during the day (except yourself) " \
                "cannot be lynched that day."

  @property
  def vote_action(self):
    return Pardon(placeholders.Self, placeholders.Other(), visible=False)

class Hitman(Role):
  description = Goon.description + " Your target cannot be protected."

  @property
  def faction_action(self):
    return HitmanKill(placeholders.Self(), placeholders.Player())

class Joker(NeutralRole):
  description = "You win only if you get lynched."

  def fate(self, *, player, game):
    if player.alive:
      return Fate.undecided
    if game.log.has_been_lynched(player):
      return Fate.won
    return Fate.lost

class Lyncher(NeutralRole):
  description = "You win only if a specific player gets lynched."

  def set_target(self, target):
    self.target = target

  def fate(self, *, player, game):
    assert self.target
    if self.target.alive:
      return Fate.undecided
    if game.log.has_been_lynched(self.target):
      return Fate.won
    return Fate.lost

class Miller(Role):
  description = "You appear guilty to cop investigations."

  @property
  def apparent_alignment(self):
    return Alignment.evil

class Ninja(ModifierRole):
  description = "Your night actions are invisible to Trackers, Watchers, " \
                "and Forensic Investigators."

  def __init__(self, faction_or_role):
    super().__init__(faction_or_role)
    assert faction_or_role.action or faction_or_role.faction_action

  @property
  def visible(self):
    return False

class Overeager(ModifierRole):
  description = "You MUST use your action each night. If you do not, " \
                "a target will be chosen for you at random."

  def __init__(self, role):
    super().__init__(role)
    assert role.action

  @property
  def action(self):
    if self.base.action:
      return Compelled(self.base.action)

class ParanoidGunOwner(Role):
  description = "You automatically kill any player who visits you."

  @property
  def kills_visitors(self):
    return True

class Politician(Role):
  description = "You may steal one player's vote each night. " \
                "That player automatically votes with you the next day."

  @property
  def action(self):
    return StealVote(placeholders.Self(), placeholders.Player())

class Roleblocker(Role):
  description = "You may roleblock one player each night. " \
                "That player cannot use their action that night."

  @property
  def action(self):
    return Roleblock(placeholders.Self(), placeholders.Player())

class Tracker(Role):
  description = "You may track one player each night. " \
                "You discover everyone they visit that night."

  @property
  def action(self):
    return Track(placeholders.Self(), placeholders.Player())

class Unlynchable(ModifierRole):
  description = "You cannot be lynched."

  def on_lynched(self, *, game, player):
    game.log.append(events.NoLynch())

class Usurper(Goon):
  description = "You win with your faction, but only if you are " \
                "its leader at the end of the game."

  def fate(self, *, player, game):
    faction_fate = self.base.fate(player=player, game=game)
    if faction_fate == Fate.won:
      return Fate.won if self.faction.leader(game=game) == player else Fate.lost
    return faction_fate

class Watcher(Role):
  description = "You may watch one player each night. " \
                "You discover everyone who visits them that night."

  @property
  def action(self):
    return Watch(placeholders.Self(), placeholders.Player())

class Vengeful(ModifierRole):
  description = "You automatically kill anyone who kills you."

  def on_killed(self, *, game, player, by, protectable, **kwargs):
    self.base.on_killed(game=game, player=player, by=by,
                        protectable=protectable, **kwargs)
    resolve_kill(player, by, game=game)

class Ventriloquist(Role):
  description = "You may possess one player each night. " \
                "You may override their target with someone of " \
                "your choosing."

  @property
  def action(self):
    return Possess(placeholders.Self(), placeholders.Player(), placeholders.Player())

class Vigilante(Role):
  description = "You may kill one player each night."

  @property
  def action(self):
    return Kill(placeholders.Self(), placeholders.Player())

class Villager(Role):
  description = "You have no special abilities."



from .actions import *
from .alignment import *
from . import placeholders
from .util import *
from .virtual_actions import *

import enum

class Fate(enum.Enum):
  won       = 1
  lost      = 2
  undecided = 3

class Faction(object):
  wins_exclusively  = True
  secret_membership = False

  def __init__(self, name):
    self.name = name

  def __lt__(self, other):
    return self.name < other.name

  def __str__(self):
    return self.name

  def leader(self, *, game):
    members = [p for p in game.player_list if p.faction == self and p.alive]
    return members[0] if len(members) > 0 else None

  def members(self, *, game):
    return [p for p in game.players if p.faction == self]

  def all_members(self, *, game):
    return [p for p in game.all_players if p.faction == self]

  def action(self, *, game):
    member_actions = [p.faction_action for p in self.members(game=game) if p.faction_action]
    if member_actions: return FactionAction(self, OneOfAction(member_actions))

class Town(Faction):
  adjective         = "Town"
  alignment         = Alignment.good
  objective         = "You win if all the mafia are eliminated."
  secret_membership = True

  def __init__(self):
    super().__init__("Town")

  def fate(self, *, game, **kwargs):
    good_players = len([p for p in game.players if p.alignment == Alignment.good])
    evil_players = len([p for p in game.players if p.alignment == Alignment.evil])
    if good_players == 0: return Fate.lost
    if evil_players == 0: return Fate.won
    return Fate.undecided

class Mafia(Faction):
  adjective = "Mafia"
  alignment = Alignment.evil
  objective = "You win if your faction outnumbers all other players in the game."

  def fate(self, *, game, **kwargs):
    members = self.members(game=game)
    if len(members) == 0: return Fate.lost
    if len(members) > len(game.players) / 2: return Fate.won
    return Fate.undecided

class Masonry(Faction):
  adjective = "Mason"
  alignment = Alignment.good
  objective = Town.objective

  def __init__(self, name, town):
    super().__init__(name)
    self.town = town

  @property
  def alignment(self):
    return self.town.alignment

  def fate(self, **kwargs):
    return self.town.fate(**kwargs)

class ThirdParty(Faction):
  adjective = "Third-Party"
  secret_membership = True



from . import events
from .exceptions import *
from .log import *
import mafia.night
from .phase import *
from .util import *

from collections import *
import re

class Day(Phase):
  """
  A single day in a Mafia game.

  Usage:
    day1 = Day(1)
    day1.set_vote(cop, goon)
    day1.set_vote(goon, cop)
    day1.set_vote(villager, cop)
    game.resolve(day1)
  """

  def __init__(self, number):
    self.number     = number
    self.votes      = defaultdict(return_none)
    self.vote_order = []

  def __eq__(self, other):
    return isinstance(other, Day) and other.number == self.number

  def __str__(self):
    return "Day %d" % self.number

  def next_phase(self):
    return mafia.Night(self.number)

  def set_vote(self, player, candidate):
    self.votes[player] = candidate

    # Keep track of voting order for vote action resolution
    if player in self.vote_order:
      self.vote_order.remove(player)
    self.vote_order.append(player)

  def _resolve(self, game):
    votes      = defaultdict(return_none)
    candidates = defaultdict(return_0)

    # Determine actual votes after politicianing
    for player in game.players:
      if player.votes_with.alive:
        votes[player] = self.votes[player.votes_with]

    # Apply vote actions
    vote_actions = []
    for player in self.vote_order:
      if player.vote_action:
        action = player.vote_action.with_target(votes[player])
        if player.vote_action.matches(action):
          vote_actions.append(action)
    self.resolve_actions(vote_actions, game=game)

    # Count votes
    for player in game.players:
      if votes[player] and votes[player].alive:
        game.log.append(events.VotedFor(player, votes[player], votes=player.votes, original_vote=self.votes[player]))
        candidates[votes[player]] += player.votes

    # Try to lynch the first viable candidate by number of votes or coin flip
    candidate_list = game.shuffled([(c,p) for p,c in sorted(candidates.items())])
    candidate_list = [p for c,p in sorted(candidate_list, reverse=True)]
    victim = candidate_list[0] if len(candidate_list) > 0 else None
    if victim:
      victim.on_lynched(game=game, player=victim)
    else:
      game.log.append(events.NoLynch())

    # Advance effects
    for player in game.all_players:
      for effect in player.effects:
        effect.expiration.advance_day()



from .exceptions import *
from .util import *

import copy

class TargetList(list):
  def matches(self, other, **kwargs):
    return len(self) == len(other) and \
           all([s.matches(o, **kwargs) for s, o in zip(self, other)])

  def random_instance(self, **kwargs):
    return TargetList([t.random_instance(**kwargs) for t in self])

class ActionBase(object):
  compelled = False

  def concrete_action(self):
    raise NotImplementedError()

  def matches(self, other, **kwargs):
    return matches(self, other, **kwargs)

  def with_player(self, player):
    clone = copy.copy(self)
    clone.player = player
    return clone

  def with_target(self, target):
    return self.with_targets([target])

  def with_targets(self, targets):
    clone = copy.copy(self)
    clone.raw_targets = TargetList(targets)
    return clone

  def random_instance(self, **kwargs):
    clone = copy.copy(self)
    fill_randomly(clone, **kwargs)
    return clone



import copy
import functools
import re

from .day import *
from .night import *
from . import placeholders

class Command(object):
  def __init__(self, pattern, command, *, help=None, phase=None):
    self.help    = help or pattern
    self.pattern = pattern
    self.command = command
    self.phase   = phase

    pattern = pattern.replace("PLAYER", r"(\w+)")
    pattern = pattern.replace("TEXT", r"(.*)")
    self.regex = re.compile(pattern, re.DOTALL + re.IGNORECASE)

# TODO: Generalize.
class Parser(object):
  def __init__(self, game):
    self.game = game
    self.commands = [
      Command(r"\s*(?:vote|lynch) PLAYER.*?", self.update_vote, phase=Day,
              help="vote PLAYER"),
      Command(r"\s*(?:leave|set|update|write) will:?TEXT", self.update_will,
              help="set will: ..."),
      Command(r"\s*help.*?", self.show_help,
              help="help"),
    ]

  def parse(self, phase, player, message):
    for command in self.get_commands(player):
      match = command.regex.fullmatch(message)
      if match:
        if command.phase and not isinstance(phase, command.phase):
          raise WrongPhase(right_phase=command.phase)
        return command.command(phase, player, *match.groups())
    raise InvalidAction()

  def get_commands(self, player, *, phase=None):
    commands = []
    if player.action:
      commands.append(self.get_action_command(player.action))
    if player.faction.leader(game=self.game) == player:
      for member in player.faction.members(game=self.game):
        if member.faction_action:
          commands.append(self.get_faction_action_command(player.faction, member.faction_action))
    if phase:
      commands = [c for c in commands if isinstance(c.phase, phase)]
    return commands + self.commands

  def get_action_command(self, action):
    targets = " ".join(["PLAYER" for t in action.targets])
    pattern = "\s*%s %s.*?" % (action.name, targets)
    help = "%s %s" % (action.name, targets)
    return Command(pattern, self.perform_action(action), phase=Night, help=help)

  def get_faction_action_command(self, faction, action):
    command = self.get_action_command(action)
    pattern = "\s*%s: %s.*?" % (action.player.unique_name, command.pattern)
    help = "%s: %s" % (action.player.unique_name, command.help)
    return Command(pattern, self.perform_faction_action(faction, action), phase=Night, help=help)

  def get_help(self, player, *, phase=None):
    return [c.help for c in self.get_commands(player, phase=phase)]

  def get_player(self, name):
    name = name.lower()
    if name == "nobody":
      raise UndoRequested()
    player = self.game.player_named(name)
    if not player:
      raise InvalidPlayer(name)
    return player

  def perform_action(self, action):
    def inner(phase, player, *targets):
      try:
        targets = [self.get_player(player) for player in targets]
        real_action = action.with_targets(targets)
        if not action.matches(real_action):
          raise IllegalAction()
        phase.add_action(real_action)
      except UndoRequested:
        targets = [placeholders.Any() for t in targets]
        match = action.with_targets(targets)
        phase.raw_actions = list(filter(lambda a: not match.matches(a), phase.raw_actions))

    return inner

  def perform_faction_action(self, faction, action):
    def inner(phase, player, *targets):
      targets = [self.get_player(player) for player in targets]
      phase.add_action(FactionAction(faction, action.with_targets(targets)))
    return inner

  def update_vote(self, phase, player, vote=None):
    try:
      vote = vote and self.get_player(vote)
    except UndoRequested:
      vote = None
    phase.set_vote(player, vote)

  def update_will(self, phase, player, will):
    player.will = will.strip()

  def show_help(self, phase, player):
    raise HelpRequested()



import re

from .action_helpers import *
from .day import *
from .night import *
from .util import *

class Expiration(object):
  """An expiration condition for an Effect."""
  def expired(self):       raise NotImplementedError()
  def advance_day(self):   pass
  def advance_night(self): pass
  def on_used(self):       pass

class Days(Expiration):
  """Expires after a number of days. Unaffected by intervening nights."""
  def __init__(self, days):
    self.days = days
  def __str__(self):
    return "Days(%d)" % self.days
  def expired(self):     return self.days <= 0
  def advance_day(self): self.days -= 1

class Manual(Expiration):
  """A manually set expiration."""
  def __init__(self):
    self._expired = False
  def set_expired(self, expired):
    self._expired = expired
  def expired(self):
    return self._expired

class Nights(Expiration):
  """Expires after number of nights. Unaffected by intervening days."""
  def __init__(self, nights):
    self.nights = nights
  def __str__(self):
    return "Nights(%d)" % self.nights
  def expired(self):       return self.nights <= 0
  def advance_night(self): self.nights -= 1

class Never(Expiration):
  """Never expires."""
  def __str__(self):
    return "Never()"
  def expired(self): return False

class Effect(object):
  """Temporary alterations which can be applied to players or roles."""

  def __init__(self, *, expiration=None):
    self.expiration = expiration or Days(1)

  def __str__(self):
    return "%s(expiration=%s)" % (self.__class__.__name__, str(self.expiration))

  def __getattribute__(self, attr):
    special = re.compile(r"__.*__|expiration")
    if special.match(attr):
      return super().__getattribute__(attr)

    if self.expiration.expired():
      raise AttributeError()

    return super().__getattribute__(attr)

class Blocked(Effect):
  @property
  def blocked(self):
    return True

class ChangeRole(Effect):
  def __init__(self, role, *, expiration=None, **kwargs):
    super().__init__(expiration=expiration or Never(), **kwargs)
    self._role = role

  @property
  def role(self):
    return self._role

class Dead(Effect):
  def __init__(self, *, expiration=None, **kwargs):
    super().__init__(expiration=expiration or Never(), **kwargs)

  @property
  def alive(self):
    return False

class Delayed(Effect):
  @property
  def delayed(self):
    return True

class ExtraAction(Effect):
  def __init__(self, extra_actions=1, *, expiration=None, **kwargs):
    super().__init__(expiration=expiration or Nights(2), **kwargs)
    self.extra_actions = extra_actions

  def action_count_fn(self, base):
    return self.extra_actions + base()

class GuardedBy(Effect):
  def __init__(self, bodyguard, *, elite=False, **kwargs):
    super().__init__(**kwargs)
    self.bodyguard = bodyguard
    self.elite     = elite

  def on_killed_fn(self, base, *, game, player, by, protectable, **kwargs):
    if protectable:
      game.log.append(events.Protected(player))
      resolve_kill(by, self.bodyguard, game=game, **kwargs)
      if self.elite:
        resolve_kill(self.bodyguard, by, game=game)
    else:
      base()

class Unlynchable(Effect):
  def on_lynched_fn(self, base, *, game, player):
    game.log.append(events.NoLynch())

class MustTarget(Effect):
  def __init__(self, must_target, **kwargs):
    super().__init__(**kwargs)
    self.must_target = must_target

class Protected(Effect):
  def on_killed_fn(self, base, *, game, player, by, protectable, **kwargs):
    if protectable: game.log.append(events.Protected(player))
    else:           return base()

class Stoned(Effect):
  def __init__(self, **kwargs):
    super().__init__(expiration=Manual(), **kwargs)

  def on_killed_fn(self, base, *, game, player, by, protectable, **kwargs):
    if protectable:
      game.log.append(events.Protected(player))
      self.expiration.set_expired(True)
    else:
      return base()

class SwitchedWith(Effect):
  def __init__(self, switched_with, **kwargs):
    super().__init__(**kwargs)
    self.switched_with = switched_with

class VotesWith(Effect):
  def __init__(self, votes_with, **kwargs):
    super().__init__(**kwargs)
    self.votes_with = votes_with



from . import events
from .util import *

from collections import *
from enum import *

class Log(list):
  def __init__(self, list=None, *, phase=None):
    if list: super().__init__(list)
    if phase:
      for event in self:
        event.phase = phase
    self.append_callbacks = []
    self.current_phase = None

  def __str__(self):
    events = "\n".join(["  %s," % event.colored_str() for event in self])
    return "Log:\n%s" % events

  def __add__(self, other):
    return Log(super().__add__(other))

  def append(self, event):
    event.phase = self.current_phase
    for f in self.append_callbacks: f(event)
    super().append(event)

  def on_append(self, callback):
    self.append_callbacks.append(callback)

  def filter(self, predicate):
    return Log(filter(predicate, self))

  def phase(self, phase):
    return self.filter(lambda event: event.phase == phase)

  def this_phase(self):
    return self.phase(self.current_phase)

  def to(self, to):
    return self.filter(lambda event: event.to == events.PUBLIC or
                                     event.to == to or
                                     (event.to and to in event.to))

  def type(self, type):
    return self.filter(lambda event: isinstance(event, type))

  def has_been_lynched(self, player):
    return len(self.type(events.Lynched).filter(lambda lynch: lynch.player == player)) > 0

  def visits_by(self, player, *, include_invisible=False):
    def include(visit):
      return visit.player == player and (visit.visible or include_invisible)
    return self.type(events.Visited).filter(include)

  def visits_to(self, player, *, include_invisible=False):
    def include(visit):
      return visit.target == player and (visit.visible or include_invisible)
    return self.type(events.Visited).filter(include)



import re

from .action_base import *
from . import placeholders
from .util import *

class VirtualAction(ActionBase):
  """An action that wraps and modifies another action."""

  def __init__(self, action):
    super().__init__()
    self.action = action

  @property
  def player(self):
    return self.action.player

  @player.setter
  def player(self, player):
    self.action = self.action.with_player(player)

  def concrete_action(self):
    return self.action.concrete_action()

class FactionAction(VirtualAction):
  """An action taken by a player on behalf of their faction."""

  def __init__(self, faction, action):
    super().__init__(action)
    self.faction = faction

  def __str__(self):
    return "FactionAction(%s, %s)" % (self.faction, self.action)

class OneOfAction(VirtualAction):
  """An action that may be one of a number of different actions."""
  def __init__(self, actions):
    self.actions = actions

  def __str__(self):
    actions = ", ".join([str(a) for a in self.actions])
    return "OneOfAction(%s)" % actions

  def matches(self, other, **kwargs):
    return any([a.matches(other, **kwargs) for a in self.actions])

  def random_instance(self, *, game, **kwargs):
    return game.random.choice(self.actions).random_instance(game=game, **kwargs)

class Compelled(VirtualAction):
  """An action that MUST be used each night."""

  compelled = True

  def __str__(self):
    return "Compelled(%s)" % self.action

  def matches(self, other, **kwargs):
    return self.action.matches(other, **kwargs)

  def random_instance(self, **kwargs):
    return self.action.random_instance(**kwargs)



import mafia.day
from .exceptions import *
from .log import *
from .phase import *
from .util import *
from .virtual_actions import *

class Night(Phase):
  """
  A single night in a Mafia game.

  Usage:
    night0 = Night(0)
    night0.add_action(Investigate(cop, goon))
    night0.add_action(FactionAction(mafia, Kill(cop, goon)))
    game.resolve(night0)
  """

  def __init__(self, number):
    self.number      = number
    self.raw_actions = []

  def __eq__(self, other):
    return isinstance(other, Night) and other.number == self.number

  def __str__(self):
    return "Night %d" % self.number

  def next_phase(self):
    return mafia.Day(self.number + 1)

  def add_action(self, action):
    self.raw_actions.append(action)

  def _resolve(self, game):
    # Compile valid action set
    options = []
    for player in game.players:
      if player.action:
        for i in range(player.action_count):
          options.append(player.action)
    for faction in game.factions:
      if faction.action(game=game):
        options.append(faction.action(game=game))

    # Check actions
    actions = []
    for action in reversed(self.raw_actions):
      for option in options:
        if option.matches(action, game=game):
          actions.append(action.concrete_action())
          options.remove(option)
          break
    actions.reverse()

    # Add compelled actions
    for option in options:
      if option.compelled:
        instance = option.random_instance(game=game)
        actions.append(instance)

    # Add delayed actions
    actions = game.delayed_actions + actions
    game.delayed_actions = []

    # Resolve actions
    self.resolve_actions(actions, game=game)

    if len(game.log.phase(self).type(events.Died)) == 0:
      game.log.append(events.NoDeaths())

    # Advance effects
    for player in game.all_players:
      for effect in player.effects:
        effect.expiration.advance_night()



from setuptools import *

setup(
  # Metadata
  name="mafia",
  version="0.8",
  author="Calder Coalson",
  author_email="caldercoalson@gmail.com",
  url="https://github.com/calder/mafia",
  description="A library for moderating games of Mafia.",
  long_description="See https://github.com/calder/mafia for documentation.",

  # Contents
  packages=find_packages(exclude=["*.test"]),

  # Dependencies
  install_requires=[
    "termcolor",
  ],
  tests_require=[
    "pytest",
  ],

  # Settings
  zip_safe=True,
)



import mafia

def resolve_death(player, *, game):
  old_faction_leader = player.faction.leader(game=game)

  # Kill the player
  player.add_effect(mafia.effects.Dead())

  # Send promotion to new faction leader if necessary
  new_faction_leader = player.faction.leader(game=game)
  if new_faction_leader != old_faction_leader and player.faction.action(game=game):
     game.log.append(mafia.events.FactionLeaderAnnouncement(player.faction, new_faction_leader))

def resolve_kill(player, target, *, game, protectable=True, stack=None):
  # Skip redundant kills
  if not target.alive:
    return

  # Prevent infinite bodyguarding loops
  stack = stack or []
  if target in stack:
    protectable = False
  stack = stack + [target]

  # Resolve the kill
  target.on_killed(game=game, player=target, by=player, protectable=protectable, stack=stack)



class HelpRequested(Exception):
  pass

class InvalidAction(Exception):
  def __str__(self):
    return "Invalid action."

class IllegalAction(InvalidAction):
  def __str__(self):
    return "Illegal action."

class InvalidPlayer(InvalidAction):
  def __init__(self, player):
    self.player = player

  def __str__(self):
    return "%r is not a valid player." % self.player

class UndoRequested(Exception):
  pass

class WrongPhase(InvalidAction):
  def __init__(self, *, right_phase):
    self.right_phase = right_phase

  def __str__(self):
    return "That action can only be used during the %s." % \
           self.right_phase.__class__.__name__
